/*
 * Copyright 2017 Arne Vajhøj.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */
 
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include <ssdef.h>
#include <descrip.h>
#include <lib$routines.h>

#include "dk_vajhoej_vms_call_VMS.h"

/*
 * 32 vs 64 bit pointer crap.
 */
#pragma pointer_size save
#pragma pointer_size 32
typedef char *char_ptr32;
#pragma pointer_size 64
typedef char *char_ptr64;
#pragma pointer_size restore

#define MAX_ITEMS 100

struct DPTR
{
	char *p;
	int plen;
	char_ptr32 p32;
};

struct DPTR *create_dptr(char *p, int plen)
{
	struct DPTR *dp;
	dp = _malloc32(sizeof(struct DPTR));
	dp->p = p;
	dp->plen = plen;
#if defined(__ia64) || defined(__x86_64)
	dp->p32 = _malloc32(dp->plen);
#ifdef DEBUG
	fprintf(stderr, "    allocated %08p\n", dp->p32);
	fprintf(stderr, "    copy %d bytes from %016p (%.8s) to %08p\n", dp->plen, dp->p, dp->p, dp->p32);
#endif
	memcpy(dp->p32, dp->p, dp->plen);
#else
	dp->p32 = p;
#endif
	return dp;
}

char *destroy_dptr(struct DPTR *dp, int copyback)
{
	char *p;
	p = dp->p;
#if defined(__ia64) || defined(__x86_64)
	if(copyback)
	{
#ifdef DEBUG
		printf("    copy back %d bytes from %08p (%.8s) to %016p\n", dp->plen, dp->p32, dp->p32, dp->p);
#endif
		memcpy(dp->p, dp->p32, dp->plen);
	}
#ifdef DEBUG
	fprintf(stderr, "    deallocate %08p\n", dp->p32);
#endif
	free(dp->p32);
#endif
	free(dp);
	return p;
}

struct item32
{
	short buflen,code;
	long bufadr,retlenadr;
};

/* create descriptor */
static struct dsc$descriptor_s create_desc(const char *buf, int len)
{
   struct dsc$descriptor_s res;
   res.dsc$w_length = (short)len;
   res.dsc$b_dtype = DSC$K_DTYPE_T;
   res.dsc$b_class = DSC$K_CLASS_S;
   res.dsc$a_pointer = (char_ptr32)buf;
   return res;
}

/* cached method id's */
static jmethodID mid_TypeWrapper_getValue;
static jmethodID mid_Short_shortValue;
static jmethodID mid_Integer_intValue;
static jmethodID mid_Long_longValue;
static jmethodID mid_Float_floatValue;
static jmethodID mid_Double_doubleValue;
static jmethodID mid_TypeWrapper_setValue_short;
static jmethodID mid_TypeWrapper_setValue_int;
static jmethodID mid_TypeWrapper_setValue_long;
static jmethodID mid_TypeWrapper_setValue_float;
static jmethodID mid_TypeWrapper_setValue_double;
static jmethodID mid_Item32_getType;
static jmethodID mid_Item32_getBuflen;
static jmethodID mid_Item32_getCode;
static jmethodID mid_Item32_getBufadr;
static jmethodID mid_Item32_getRetlenadrorflag;

#ifdef DEBUG
static char clsnambuf[1024];

static char *get_class_name(JNIEnv *cntx, jobject obj) {
    jclass cls;
    jmethodID mid;
    jobject clsobj;
    jclass clscls;
    jmethodID clsmid;
    jstring str;
    const char *p;
    cls = (*cntx)->GetObjectClass(cntx, obj);
    mid = (*cntx)->GetMethodID(cntx, cls, "getClass", "()Ljava/lang/Class;");
    clsobj = (*cntx)->CallObjectMethod(cntx, obj, mid);
    clscls = (*cntx)->GetObjectClass(cntx, clsobj);
    clsmid = (*cntx)->GetMethodID(cntx, clscls, "getName", "()Ljava/lang/String;");
    str = (jstring)(*cntx)->CallObjectMethod(cntx, clsobj, clsmid);
    p = (*cntx)->GetStringUTFChars(cntx, str, NULL);
    strcpy(clsnambuf, p);
    (*cntx)->ReleaseStringUTFChars(cntx, str, p);
    return clsnambuf;
}
#endif

static jobject rbuf_ba;
static struct DPTR *rbuf_ba_dptr;

/* ROUTINE0 support (LBR$GET_HELP) */
static long routine0(struct dsc$descriptor_s *a1, long *a2, long *a3, long a4)
{
    int ix;
    ix = strlen(rbuf_ba_dptr->p32);
    if(ix + a1->dsc$w_length + 2 > rbuf_ba_dptr->plen) return SS$_ABORT;
    memcpy(rbuf_ba_dptr->p32 + ix, a1->dsc$a_pointer, a1->dsc$w_length);
    ix += a1->dsc$w_length;
    rbuf_ba_dptr->p32[ix] = '\0';
    strcat(rbuf_ba_dptr->p32, "\r\n");
    ix += 2;
    return SS$_NORMAL;
}

/* get short/int/long/float/double/descriptor from Short/Integer/Long/Float/Double/byte[] */
static short get_w(JNIEnv *cntx, jobject sobj)
{
   return (*cntx)->CallShortMethod(cntx, sobj, mid_Short_shortValue);
}

static long get_lw(JNIEnv *cntx, jobject iobj)
{
   return (*cntx)->CallIntMethod(cntx, iobj, mid_Integer_intValue);
}

static long long get_qw(JNIEnv *cntx, jobject lobj)
{
   return (*cntx)->CallLongMethod(cntx, lobj, mid_Long_longValue);
}

static float get_sf(JNIEnv *cntx, jobject fobj)
{
   return (*cntx)->CallFloatMethod(cntx, fobj, mid_Float_floatValue);
}

static double get_tf(JNIEnv *cntx, jobject dobj)
{
   return (*cntx)->CallDoubleMethod(cntx, dobj, mid_Double_doubleValue);
}

static struct DPTR *get_ba(JNIEnv *cntx, jobject baobj)
{
   signed char *buf;
   int len;
   len = (*cntx)->GetArrayLength(cntx, baobj);
   buf = (*cntx)->GetByteArrayElements(cntx, baobj, NULL);
#ifdef DEBUG
   fprintf(stderr, "    GBAE on %016p returned %016p\n", baobj, buf);
#endif
   return create_dptr((char *)buf, len);
}

/*
static int get_ba_length(JNIEnv *cntx, jobject baobj)
{
   return (*cntx)->GetArrayLength(cntx, baobj);
}
*/

/* get Object from TypeWrapper */
static jobject get_wrapped(JNIEnv *cntx, jobject twobj)
{
   return (*cntx)->CallObjectMethod(cntx, twobj, mid_TypeWrapper_getValue);
}

/* get short/int/long/float/double/byte[] from TypeWrapper */
static short get_wrapped_w(JNIEnv *cntx, jobject twobj)
{
#ifdef DEBUG
   fprintf(stderr, "    unwrap w from %s\n", get_class_name(cntx, twobj));
#endif
   return get_w(cntx, get_wrapped(cntx, twobj));
}

static long get_wrapped_lw(JNIEnv *cntx, jobject twobj)
{
#ifdef DEBUG
   fprintf(stderr, "    unwrap lw from %s\n", get_class_name(cntx, twobj));
#endif
   return get_lw(cntx, get_wrapped(cntx, twobj));
}

static long long get_wrapped_qw(JNIEnv *cntx, jobject twobj)
{
#ifdef DEBUG
   fprintf(stderr, "    unwrap qw from %s\n", get_class_name(cntx, twobj));
#endif
   return get_qw(cntx, get_wrapped(cntx, twobj));
}

static float get_wrapped_sf(JNIEnv *cntx, jobject twobj)
{
#ifdef DEBUG
   fprintf(stderr, "    unwrap sf from %s\n", get_class_name(cntx, twobj));
#endif
   return get_sf(cntx, get_wrapped(cntx, twobj));
}

static double get_wrapped_tf(JNIEnv *cntx, jobject twobj)
{
#ifdef DEBUG
   fprintf(stderr, "    unwrap tf from %s\n", get_class_name(cntx, twobj));
#endif
   return get_tf(cntx, get_wrapped(cntx, twobj));
}

/*
static struct DPTR *get_wrapped_ba(JNIEnv *cntx, jobject twobj)
{
#ifdef DEBUG
   fprintf(stderr, "    unwrap ba from %s\n", get_class_name(cntx, twobj));
#endif
   return get_ba(cntx, get_wrapped(cntx, twobj));
}

static int get_wrapped_ba_length(JNIEnv *cntx, jobject twobj)
{
#ifdef DEBUG
   fprintf(stderr, "get ba length from %s\n", get_class_name(cntx, twobj));
#endif
   return get_ba_length(cntx, get_wrapped(cntx, twobj));
}
*/

/* set short/int/long/float/double/byte[] in TypeWrapper */
static void set_wrapped_w(JNIEnv *cntx, jobject twobj, short v)
{
#ifdef DEBUG
   fprintf(stderr, "    wrap w into %s\n", get_class_name(cntx, twobj));
#endif
   (*cntx)->CallObjectMethod(cntx, twobj, mid_TypeWrapper_setValue_short, v);
}

static void set_wrapped_lw(JNIEnv *cntx, jobject twobj, long v)
{
#ifdef DEBUG
   fprintf(stderr, "    wrap lw into %s\n", get_class_name(cntx, twobj));
#endif
   (*cntx)->CallObjectMethod(cntx, twobj, mid_TypeWrapper_setValue_int, v);
}

static void set_wrapped_qw(JNIEnv *cntx, jobject twobj, long long v)
{
#ifdef DEBUG
   fprintf(stderr, "    wrap qw into %s\n", get_class_name(cntx, twobj));
#endif
   (*cntx)->CallObjectMethod(cntx, twobj, mid_TypeWrapper_setValue_long, v);
}

static void set_wrapped_sf(JNIEnv *cntx, jobject twobj, float v)
{
#ifdef DEBUG
   fprintf(stderr, "    wrap sf into %s\n", get_class_name(cntx, twobj));
#endif
   (*cntx)->CallObjectMethod(cntx, twobj, mid_TypeWrapper_setValue_float, v);
}

static void set_wrapped_tf(JNIEnv *cntx, jobject twobj, double v)
{
#ifdef DEBUG
   fprintf(stderr, "    wrap tf into %s\n", get_class_name(cntx, twobj));
#endif
   (*cntx)->CallObjectMethod(cntx, twobj, mid_TypeWrapper_setValue_double, v);
}

static void set_ba(JNIEnv *cntx, jobject baobj, struct DPTR *v, int wrtbck)
{
#ifdef DEBUG
   fprintf(stderr, "    RBAE on %016p and %016p\n", baobj, v->p);
#endif
   (*cntx)->ReleaseByteArrayElements(cntx, baobj, (signed char *)destroy_dptr(v, wrtbck), wrtbck ? 0 : JNI_ABORT);
}

/*
static void set_wrapped_ba(JNIEnv *cntx, jobject twobj, struct DPTR *v, int wrtbck)
{
#ifdef DEBUG
   fprintf(stderr, "    wrap ba into %s\n", get_class_name(cntx, twobj));
#endif
	set_ba(cntx, get_wrapped(cntx, twobj), v, wrtbck);
}
*/

/* get type from Item32 */
static long get_item_type(JNIEnv *cntx, jobject lstobj)
{
     return (*cntx)->CallIntMethod(cntx, lstobj, mid_Item32_getType);
}

static long get_item_buflen(JNIEnv *cntx, jobject lstobj)
{
     return (*cntx)->CallIntMethod(cntx, lstobj, mid_Item32_getBuflen);
}

static long get_item_code(JNIEnv *cntx, jobject lstobj)
{
     return (*cntx)->CallIntMethod(cntx, lstobj, mid_Item32_getCode);
}

static jobject get_item_bufadr(JNIEnv *cntx, jobject lstobj)
{
     return (*cntx)->CallObjectMethod(cntx, lstobj, mid_Item32_getBufadr);
}

static jobject get_item_retlenadrorflag(JNIEnv *cntx, jobject lstobj)
{
     return (*cntx)->CallObjectMethod(cntx, lstobj, mid_Item32_getRetlenadrorflag);
}

/* load cache of method id's */
JNIEXPORT jboolean JNICALL Java_dk_vajhoej_vms_call_VMS_setup(JNIEnv *cntx, jclass me)
{
   jclass cls;
   /* TypeWrapper.getValue */
   cls = (*cntx)->FindClass(cntx, "dk/vajhoej/vms/call/VMS$TypeWrapper");
   if(cls == 0) return 0;
   mid_TypeWrapper_getValue = (*cntx)->GetMethodID(cntx, cls, "getValue", "()Ljava/lang/Object;");
   if(mid_TypeWrapper_getValue == 0) return 0;
   /* Short.shortValue/Integer.intValue/Long.longValue */
   cls = (*cntx)->FindClass(cntx, "java/lang/Short");
   if(cls == 0) return 0;
   mid_Short_shortValue = (*cntx)->GetMethodID(cntx, cls, "shortValue", "()S");
   if(mid_Short_shortValue == 0) return 0;
   cls = (*cntx)->FindClass(cntx, "java/lang/Integer");
   if(cls == 0) return 0;
   mid_Integer_intValue = (*cntx)->GetMethodID(cntx, cls, "intValue", "()I");
   if(mid_Integer_intValue == 0) return 0;
   cls = (*cntx)->FindClass(cntx, "java/lang/Long");
   if(cls == 0) return 0;
   mid_Long_longValue = (*cntx)->GetMethodID(cntx, cls, "longValue", "()J");
   if(mid_Long_longValue == 0) return 0;
   cls = (*cntx)->FindClass(cntx, "java/lang/Float");
   if(cls == 0) return 0;
   mid_Float_floatValue = (*cntx)->GetMethodID(cntx, cls, "floatValue", "()F");
   if(mid_Float_floatValue == 0) return 0;
   cls = (*cntx)->FindClass(cntx, "java/lang/Double");
   if(cls == 0) return 0;
   mid_Double_doubleValue = (*cntx)->GetMethodID(cntx, cls, "doubleValue", "()D");
   if(mid_Double_doubleValue == 0) return 0;
   /* Word.setValue/LongWord.setValue/QuadWord.setValue */
   cls = (*cntx)->FindClass(cntx, "dk/vajhoej/vms/call/VMS$Word");
   if(cls == 0) return 0;
   mid_TypeWrapper_setValue_short = (*cntx)->GetMethodID(cntx, cls, "setValue", "(S)V");
   if(mid_TypeWrapper_setValue_short == 0) return 0;
   cls = (*cntx)->FindClass(cntx, "dk/vajhoej/vms/call/VMS$LongWord");
   if(cls == 0) return 0;
   mid_TypeWrapper_setValue_int = (*cntx)->GetMethodID(cntx, cls, "setValue", "(I)V");
   if(mid_TypeWrapper_setValue_int == 0) return 0;
   cls = (*cntx)->FindClass(cntx, "dk/vajhoej/vms/call/VMS$QuadWord");
   if(cls == 0) return 0;
   mid_TypeWrapper_setValue_long = (*cntx)->GetMethodID(cntx, cls, "setValue", "(J)V");
   if(mid_TypeWrapper_setValue_long == 0) return 0;
   cls = (*cntx)->FindClass(cntx, "dk/vajhoej/vms/call/VMS$SFloat");
   if(cls == 0) return 0;
   mid_TypeWrapper_setValue_float = (*cntx)->GetMethodID(cntx, cls, "setValue", "(F)V");
   if(mid_TypeWrapper_setValue_float == 0) return 0;
   cls = (*cntx)->FindClass(cntx, "dk/vajhoej/vms/call/VMS$TFloat");
   if(cls == 0) return 0;
   mid_TypeWrapper_setValue_double = (*cntx)->GetMethodID(cntx, cls, "setValue", "(D)V");
   if(mid_TypeWrapper_setValue_double == 0) return 0;
   /* Item32.getXxxx */
   cls = (*cntx)->FindClass(cntx, "dk/vajhoej/vms/call/VMS$Item32");
   if(cls == 0) return 0;
   mid_Item32_getType = (*cntx)->GetMethodID(cntx, cls, "getType", "()I");
   if(mid_Item32_getType == 0) return 0;
   mid_Item32_getBuflen = (*cntx)->GetMethodID(cntx, cls, "getBuflen", "()I");
   if(mid_Item32_getBuflen == 0) return 0;
   mid_Item32_getCode = (*cntx)->GetMethodID(cntx, cls, "getCode", "()I");
   if(mid_Item32_getCode == 0) return 0;
   mid_Item32_getBufadr = (*cntx)->GetMethodID(cntx, cls, "getBufadr", "()Ljava/lang/Object;");
   if(mid_Item32_getBufadr == 0) return 0;
   mid_Item32_getRetlenadrorflag = (*cntx)->GetMethodID(cntx, cls, "getRetlenadrorflag", "()Ljava/lang/Object;");
   if(mid_Item32_getRetlenadrorflag == 0) return 0;
   return 1;
}

/* find entrypoint in shareable image */
JNIEXPORT jlong JNICALL Java_dk_vajhoej_vms_call_VMS_getFunction(JNIEnv *cntx, jclass me, jstring image, jstring function)
{
   int result;
   int f;
   const char *image2;
   const char *function2;
   char _image[1024]; // buffer that will work with 32 bit pointer
   char _function[256]; // buffer that will work with 32 bit pointer
   struct dsc$descriptor_s imagedesc;
   struct dsc$descriptor_s functiondesc;
   image2 = (*cntx)->GetStringUTFChars(cntx, image, 0);
   strcpy(_image, image2);
   imagedesc = create_desc(_image, strlen(_image));
   function2 = (*cntx)->GetStringUTFChars(cntx, function, 0);
   strcpy(_function, function2);
   functiondesc = create_desc(_function, strlen(_function));
   f = 0;
   result = lib$find_image_symbol(&imagedesc, &functiondesc, &f);
#ifdef DEBUG
   fprintf(stderr, "LIB$FIND_IMAGE_SYMBOL image=%s function=%s result=%d f=%ld\n", image2, function2, result, f);
#endif
   (*cntx)->ReleaseStringUTFChars(cntx, image, image2);
   (*cntx)->ReleaseStringUTFChars(cntx, function, function2);
   return f;
}

/* call */
JNIEXPORT jint JNICALL Java_dk_vajhoej_vms_call_VMS_callFunction(JNIEnv *cntx, jclass me, jlong function, jintArray argtyp, jintArray argacc, jobjectArray argval)
{
   jint result; /* result code from LIB$CALLG returned ti Java */
   int narg; /* number arguments in call */
   int nitem; /* number items in item list */
   int *arglist; /* argument list for LIB$CALLG */
   jint *argtyp_int; /* argument type as C int */
   jint *argacc_int; /* argument access as C int */
   jobject *argval_tw; /* argument value as Java TypeWrapper */
   jobject *argval_ba; /* argument value for byte array as Java byte[] */
   struct DPTR **argval_ba_dptr; /* argument value for byte array as double pointer (native + 32 bit) */
   jobject *argval_itmlst_arritm; /* argument value for item list as Java array of items */
   int *temp; /* 8 byte temporary storage for by reference & descriptor for arguments */
   jobject *itmlst_item; /* item list items as Java item objects */
   int *itmlst_typ; /* item list types as C int */
   jobject *itmlst_item_ba; /* /* item list items for byte array as Java byte[] */
   struct DPTR **itmlst_item_ba_dptr; /* item list items for byte array as double pointer (native + 32 bit) */
   struct item32 *itmlst_arrstruct; /* item list as C array of struct */
   int *temp2; /*  8 byte temporary storage for by reference & descriptor for items */
   /* temporary variables only used within case label: */
   long long temp8;
   float tempf4;
   double tempf8;
   struct dsc$descriptor_s tempd;
   /* temporary pointer only used within case label: */
   int *p;
   /* multi block: */
   int *mb_ptr;
   int mb_hdrlen;
   /* loop counters: */
   int i, j, k;
   /* setup */
   narg = (*cntx)->GetArrayLength(cntx, argtyp);
   arglist = _calloc32(1 + narg, sizeof(int));
   argval_tw = calloc(narg, sizeof(jobject));
   argval_ba = calloc(narg, sizeof(jobject));
   argval_ba_dptr = _calloc32(narg, sizeof(struct DPTR *));
   argval_itmlst_arritm = calloc(narg, sizeof(jobject));
   itmlst_item = calloc(MAX_ITEMS, sizeof(jobject));
   itmlst_typ = calloc(MAX_ITEMS, sizeof(int));
   temp = _calloc32(narg * 2, sizeof(int));
   temp2 = _calloc32(MAX_ITEMS * 2, sizeof(int));
   itmlst_item_ba = _calloc32(MAX_ITEMS, sizeof(jobject));
   itmlst_item_ba_dptr = _calloc32(MAX_ITEMS, sizeof(struct DPTR *));
   itmlst_arrstruct = _calloc32(MAX_ITEMS, sizeof(struct item32));
#ifdef DEBUG
   fprintf(stderr, "call function=%lld\n", function);
#endif
   arglist[0] = narg;
   argtyp_int = (*cntx)->GetIntArrayElements(cntx, argtyp, 0);
   argacc_int = (*cntx)->GetIntArrayElements(cntx, argacc, 0);
   /* process arguments before calling native function */
   for(i = 0; i < narg; i++) 
   {
      argval_tw[i] = (*cntx)->GetObjectArrayElement(cntx, argval, i);
#ifdef DEBUG
       fprintf(stderr, "  arg=%d typ=%d acc=%d cls=%s ptr=%016p\n", i+1, argtyp_int[i], argacc_int[i], get_class_name(cntx, argval_tw[i]), argval_tw[i]);
#endif
      switch(argtyp_int[i])
      {
         case dk_vajhoej_vms_call_VMS_WORD_BYVALUE:
            arglist[1 + i] = get_wrapped_w(cntx,argval_tw[i]);
#ifdef DEBUG
            fprintf(stderr, "    wbv=%d\n", arglist[1 + i]);
#endif
            break;
         case dk_vajhoej_vms_call_VMS_LONGWORD_BYVALUE:
            arglist[1 + i] = get_wrapped_lw(cntx,argval_tw[i]);
#ifdef DEBUG
            fprintf(stderr, "    lwbv=%d\n", arglist[1 + i]);
#endif
            break;
         case dk_vajhoej_vms_call_VMS_WORD_BYREFERENCE:
            if(argacc_int[i] != dk_vajhoej_vms_call_VMS_WRITEONLY) {
                temp[2*i] = get_wrapped_w(cntx,argval_tw[i]);
            }
            arglist[i + 1] = (int)&temp[2*i];
#ifdef DEBUG
            fprintf(stderr, "    wbr=%d\n", temp[2*i]);
#endif
            break;
         case dk_vajhoej_vms_call_VMS_LONGWORD_BYREFERENCE:
            if(argacc_int[i] != dk_vajhoej_vms_call_VMS_WRITEONLY) {
                temp[2*i] = get_wrapped_lw(cntx,argval_tw[i]);
            }
            arglist[i + 1] = (int)&temp[2*i];
#ifdef DEBUG
            fprintf(stderr, "    lwbr=%d\n", temp[2*i]);
#endif
            break;                    
         case dk_vajhoej_vms_call_VMS_QUADWORD_BYREFERENCE:
            if(argacc_int[i] != dk_vajhoej_vms_call_VMS_WRITEONLY) {
                temp8 = get_wrapped_qw(cntx,argval_tw[i]);
                memcpy(&temp[2*i], &temp8, sizeof(long long));
            }
            arglist[i + 1] = (int)&temp[2*i];
#ifdef DEBUG
            fprintf(stderr, "    qwbr=%lld\n", temp8);
#endif
            break;
         case dk_vajhoej_vms_call_VMS_SFLOAT_BYREFERENCE:
            if(argacc_int[i] != dk_vajhoej_vms_call_VMS_WRITEONLY) {
                tempf4 = get_wrapped_sf(cntx,argval_tw[i]);
                memcpy(&temp[2*i], &tempf4, sizeof(float));
            }
            arglist[i + 1] = (int)&temp[2*i];
#ifdef DEBUG
            fprintf(stderr, "    sfbr=%f\n", tempf4);
#endif
            break;
         case dk_vajhoej_vms_call_VMS_TFLOAT_BYREFERENCE:
            if(argacc_int[i] != dk_vajhoej_vms_call_VMS_WRITEONLY) {
                tempf8 = get_wrapped_tf(cntx,argval_tw[i]);
                memcpy(&temp[2*i], &tempf8, sizeof(double));
            }
            arglist[i + 1] = (int)&temp[2*i];
#ifdef DEBUG
            fprintf(stderr, "    tfbr=%f\n", tempf8);
#endif
            break;
         case dk_vajhoej_vms_call_VMS_CHARACTERSTRING_BYDESCRIPTOR:
         	argval_ba[i] = get_wrapped(cntx,argval_tw[i]);
        	argval_ba_dptr[i] = get_ba(cntx,argval_ba[i]);
            tempd = create_desc(argval_ba_dptr[i]->p32, argval_ba_dptr[i]->plen);
            memcpy(&temp[2*i], &tempd, sizeof(struct dsc$descriptor_s));
            arglist[i + 1] = (int)&temp[2*i];
#ifdef DEBUG
            p = (int *)&tempd;
            fprintf(stderr, "    csbd=%08X %08X=%.8s (%d)\n", p[0], p[1], tempd.dsc$a_pointer, tempd.dsc$w_length);
#endif
            break;
         case dk_vajhoej_vms_call_VMS_BYTEBUFFER_BYREFERENCE:
           	 argval_ba[i] = get_wrapped(cntx,argval_tw[i]);
         	 argval_ba_dptr[i] = get_ba(cntx,argval_ba[i]);
             arglist[i + 1] = (int)argval_ba_dptr[i]->p32;
#ifdef DEBUG
             p = (int *)arglist[i + 1];
             fprintf(stderr, "    bbbr=%08X %08X\n", p[0], p[1]);
#endif
             break;
         case dk_vajhoej_vms_call_VMS_BLOCK_BYREFERENCE:
           	 argval_ba[i] = get_wrapped(cntx,argval_tw[i]);
         	 argval_ba_dptr[i] = get_ba(cntx,argval_ba[i]);
             arglist[i + 1] = (int)argval_ba_dptr[i]->p32;
#ifdef DEBUG
             p = (int *)arglist[i + 1];
             fprintf(stderr, "    bbr=%08X %08X\n", p[0], p[1]);
#endif
             break;
         case dk_vajhoej_vms_call_VMS_MULTIBLOCK_BYREFERENCE:
           	 argval_ba[i] = get_wrapped(cntx,argval_tw[i]);
         	 argval_ba_dptr[i] = get_ba(cntx,argval_ba[i]);
             mb_ptr = (int *)argval_ba_dptr[i]->p32;
             mb_hdrlen = 1 + *mb_ptr;
             for(k = 0; k < mb_hdrlen / 2; k++)
             {
#ifdef DEBUG
            	 fprintf(stderr, "    update pointer ix=%d offs=%d value=%08X\n", mb_ptr[1+2*k], mb_ptr[1+2*k+1], *(mb_ptr + mb_hdrlen + mb_ptr[1+2*k+1]/4));
#endif
                 mb_ptr[mb_hdrlen + mb_ptr[1+2*k]/4] = (int)(mb_ptr + mb_hdrlen + mb_ptr[1+2*k+1]/4);
             }
             arglist[i + 1] = (int)(mb_ptr + mb_hdrlen);
#ifdef DEBUG
             p = (int *)arglist[i + 1];
             fprintf(stderr, "    mbbr=%08X %08X %08X %08X (mb_hdrlen=%d)\n", p[0], p[1], p[2], p[3], mb_hdrlen);
#endif
             break;
         case dk_vajhoej_vms_call_VMS_ITEMLIST32_BYREFERENCE:
             argval_itmlst_arritm[i] = get_wrapped(cntx, argval_tw[i]);
             nitem = (*cntx)->GetArrayLength(cntx, argval_itmlst_arritm[i]);
             /* process item list */
             for(j = 0; j < nitem; j++)
             {
                 itmlst_item[j] = (*cntx)->GetObjectArrayElement(cntx, argval_itmlst_arritm[i], j);
                 itmlst_typ[j] = get_item_type(cntx, itmlst_item[j]);
#ifdef DEBUG
                 fprintf(stderr, "  item=%d typ=%d cls=%s\n", j+1, itmlst_typ[j], get_class_name(cntx, itmlst_item[j]));
#endif
                 itmlst_arrstruct[j].buflen = get_item_buflen(cntx, itmlst_item[j]);
                 itmlst_arrstruct[j].code = get_item_code(cntx, itmlst_item[j]);
                 switch(itmlst_typ[j]) {
                     case dk_vajhoej_vms_call_VMS_ITEM_LONGWORD_IN:
                         temp2[2*j] = get_wrapped_lw(cntx, get_item_bufadr(cntx, itmlst_item[j]));
                         itmlst_arrstruct[j].bufadr = (long)&temp2[2*j];
                         itmlst_arrstruct[j].retlenadr = get_lw(cntx, get_item_retlenadrorflag(cntx, itmlst_item[j]));
                         break;
                     case dk_vajhoej_vms_call_VMS_ITEM_LONGWORD_OUT:
                         temp2[2*j] = get_wrapped_lw(cntx, get_item_bufadr(cntx, itmlst_item[j]));
                         itmlst_arrstruct[j].bufadr = (long)&temp2[2*j];
                         itmlst_arrstruct[j].retlenadr = 0;
                         break;
                     case dk_vajhoej_vms_call_VMS_ITEM_QUADWORD_IN:
                         temp8 = get_wrapped_qw(cntx, get_item_bufadr(cntx, itmlst_item[j]));
                         memcpy(&temp2[2*j], &temp8, sizeof(long long));
                         itmlst_arrstruct[j].bufadr = (long)&temp2[2*j];
                         itmlst_arrstruct[j].retlenadr = get_lw(cntx, get_item_retlenadrorflag(cntx, itmlst_item[j]));
                         break;
                     case dk_vajhoej_vms_call_VMS_ITEM_QUADWORD_OUT:
                         temp8 = get_wrapped_qw(cntx, get_item_bufadr(cntx, itmlst_item[j]));
                         memcpy(&temp2[2*j], &temp8, sizeof(long long));
                         itmlst_arrstruct[j].bufadr = (long)&temp2[2*j];
                         itmlst_arrstruct[j].retlenadr = 0;
                         break;
                     case dk_vajhoej_vms_call_VMS_ITEM_CHARACTERSTRING_IN:
                    	 itmlst_item_ba[j] = get_wrapped(cntx,get_item_bufadr(cntx,itmlst_item[j]));
                    	 itmlst_item_ba_dptr[j] = get_ba(cntx,itmlst_item_ba[j]);
                         tempd = create_desc(itmlst_item_ba_dptr[j]->p32, itmlst_item_ba_dptr[j]->plen);
                         memcpy(&temp2[2*j], &tempd, sizeof(struct dsc$descriptor_s));
                         itmlst_arrstruct[j].bufadr = temp2[2*j+1];
                         itmlst_arrstruct[j].retlenadr = get_lw(cntx, get_item_retlenadrorflag(cntx, itmlst_item[j]));
                         break;
                     case dk_vajhoej_vms_call_VMS_ITEM_CHARACTERSTRING_OUT:
                    	 itmlst_item_ba[j] = get_wrapped(cntx,get_item_bufadr(cntx,itmlst_item[j]));
                    	 itmlst_item_ba_dptr[j] = get_ba(cntx,itmlst_item_ba[j]);
                         tempd = create_desc(itmlst_item_ba_dptr[j]->p32, itmlst_item_ba_dptr[j]->plen);
                         memcpy(&temp2[2*j], &tempd, sizeof(struct dsc$descriptor_s));
                         itmlst_arrstruct[j].bufadr = temp2[2*j+1];
                         temp2[2*j] = get_wrapped_lw(cntx, get_item_retlenadrorflag(cntx, itmlst_item[j]));
                         itmlst_arrstruct[j].retlenadr = (long)&temp2[2*j];
                         break;
                 }
             }
             itmlst_arrstruct[nitem].buflen = 0;
             itmlst_arrstruct[nitem].code = 0;
             itmlst_arrstruct[nitem].bufadr = 0;
             itmlst_arrstruct[nitem].retlenadr = 0;
             arglist[i + 1] = (int)&itmlst_arrstruct[0];
#ifdef DEBUG
             for(j = 0; j <= nitem; j++)
             {
                 fprintf(stderr, "  %04X%04X\n", itmlst_arrstruct[j].code, itmlst_arrstruct[j].buflen);
                 if(itmlst_arrstruct[j].bufadr > 0x1000)
                 {
                     p = (int *)itmlst_arrstruct[j].bufadr;
                     fprintf(stderr, "  %08X -> %08X%08X\n", itmlst_arrstruct[j].bufadr,  p[0], p[1]);
                 }
                 else
                 {
                     fprintf(stderr, "  %08X\n", itmlst_arrstruct[j].bufadr);
                 }
                 fprintf(stderr, "  %08X\n", itmlst_arrstruct[j].retlenadr);
             }
#endif
             break;
         case dk_vajhoej_vms_call_VMS_ROUTINE_0:
             rbuf_ba = get_wrapped(cntx,argval_tw[i]);
             rbuf_ba_dptr = get_ba(cntx,rbuf_ba);
             arglist[i + 1] = (int)&routine0;
 #ifdef DEBUG
             fprintf(stderr, "    r=%d rbuflen=%d\n", arglist[i + 1], rbuf_ba_dptr->plen);
 #endif
             break;
      }
   }
#ifdef DEBUG
   fprintf(stderr, "  before call:\n");
   fprintf(stderr, "  narg=%d\n", narg);
   for(i = 0; i < narg; i++)
   {
	   fprintf(stderr, "  arg[%d]=%p\n", 1 + i, (void *)arglist[1 + i]);
   }
#endif
   /* actual call of native function */
   result = lib$callg(arglist, function);
#ifdef DEBUG
   fprintf(stderr, "  sts=%d\n", result);
   fprintf(stderr, "  after call:\n");
   fprintf(stderr, "  narg=%d\n", narg);
   for(i = 0; i < narg; i++)
   {
	   fprintf(stderr, "  arg[%d]=%p\n", 1 + i, (void *)arglist[1 + i]);
   }
#endif
   /* process arguments after calling native function */
   for(i = 0; i < narg; i++) 
   {
#ifdef DEBUG
       fprintf(stderr, "  arg=%d typ=%d acc=%d cls=%s ptr=%016p\n", i+1, argtyp_int[i], argacc_int[i], get_class_name(cntx, argval_tw[i]), argval_tw[i]);
#endif
      switch(argtyp_int[i])
      {
         case dk_vajhoej_vms_call_VMS_WORD_BYVALUE:
            /* nothing */
            break;
         case dk_vajhoej_vms_call_VMS_LONGWORD_BYVALUE:
            /* nothing */
            break;
         case dk_vajhoej_vms_call_VMS_WORD_BYREFERENCE:
            if(argacc_int[i] != dk_vajhoej_vms_call_VMS_READONLY) {
                set_wrapped_w(cntx,argval_tw[i],temp[2*i]);
            }
#ifdef DEBUG
            fprintf(stderr, "    wbr=%d\n", temp[2*i]);
#endif
            break;
         case dk_vajhoej_vms_call_VMS_LONGWORD_BYREFERENCE:
            if(argacc_int[i] != dk_vajhoej_vms_call_VMS_READONLY) {
                set_wrapped_lw(cntx,argval_tw[i],temp[2*i]);
            }
#ifdef DEBUG
            fprintf(stderr, "    lwbr=%d\n", temp[2*i]);
#endif
            break;                    
         case dk_vajhoej_vms_call_VMS_QUADWORD_BYREFERENCE:
            if(argacc_int[i] != dk_vajhoej_vms_call_VMS_READONLY) {
                memcpy(&temp8,&temp[2*i],sizeof(long long));
                set_wrapped_qw(cntx,argval_tw[i],temp8);
            }
#ifdef DEBUG
            fprintf(stderr, "    qwbr=%lld\n", temp8);
#endif
            break;
         case dk_vajhoej_vms_call_VMS_SFLOAT_BYREFERENCE:
            if(argacc_int[i] != dk_vajhoej_vms_call_VMS_READONLY) {
                memcpy(&tempf4,&temp[2*i],sizeof(float));
                set_wrapped_sf(cntx,argval_tw[i],tempf4);
            }
#ifdef DEBUG
            fprintf(stderr, "    sfbr=%f\n", tempf4);
#endif
            break;
         case dk_vajhoej_vms_call_VMS_TFLOAT_BYREFERENCE:
            if(argacc_int[i] != dk_vajhoej_vms_call_VMS_READONLY) {
                memcpy(&tempf8,&temp[2*i],sizeof(double));
                set_wrapped_tf(cntx,argval_tw[i],tempf8);
            }
#ifdef DEBUG
            fprintf(stderr, "    tfbr=%f\n", tempf8);
#endif
            break;
         case dk_vajhoej_vms_call_VMS_CHARACTERSTRING_BYDESCRIPTOR:
#ifdef DEBUG
            memcpy(&tempd, &temp[2*i], sizeof(struct dsc$descriptor_s));
            p = (int *)&tempd;
            fprintf(stderr, "    csbd=%08X %08X=%.8s (%d)\n", p[0], p[1], tempd.dsc$a_pointer, tempd.dsc$w_length);
#endif
            set_ba(cntx,argval_ba[i],argval_ba_dptr[i],argacc_int[i] != dk_vajhoej_vms_call_VMS_READONLY);
            break;
         case dk_vajhoej_vms_call_VMS_BYTEBUFFER_BYREFERENCE:
#ifdef DEBUG
             p = (int *)arglist[i + 1];
             fprintf(stderr, "    bbbr=%08X %08X\n", p[0], p[1]);
#endif
             set_ba(cntx,argval_ba[i],argval_ba_dptr[i],argacc_int[i] != dk_vajhoej_vms_call_VMS_READONLY);
             break;
         case dk_vajhoej_vms_call_VMS_BLOCK_BYREFERENCE:
#ifdef DEBUG
             p = (int *)arglist[i + 1];
             fprintf(stderr, "    bbr=%08X %08X\n", p[0], p[1]);
#endif
             set_ba(cntx,argval_ba[i],argval_ba_dptr[i],argacc_int[i] != dk_vajhoej_vms_call_VMS_READONLY);
             break;
         case dk_vajhoej_vms_call_VMS_MULTIBLOCK_BYREFERENCE:
#ifdef DEBUG
             p = (int *)arglist[i + 1];
             fprintf(stderr, "    mbbr=%08X %08X %08X %08X (mb_hdrlen=%d)\n", p[0], p[1], p[2], p[3], mb_hdrlen);
#endif
             set_ba(cntx,argval_ba[i],argval_ba_dptr[i],argacc_int[i] != dk_vajhoej_vms_call_VMS_READONLY);
             break;
         case dk_vajhoej_vms_call_VMS_ITEMLIST32_BYREFERENCE:
             /* process item list */
             for(j = 0; j < nitem; j++)
             {
#ifdef DEBUG
                 fprintf(stderr, "  item=%d typ=%d cls=%s\n", j+1, itmlst_typ[j], get_class_name(cntx, itmlst_item[j]));
#endif
                 switch(itmlst_typ[j]) {
                     case dk_vajhoej_vms_call_VMS_ITEM_LONGWORD_IN:
                         /* nothing */
                         break;
                     case dk_vajhoej_vms_call_VMS_ITEM_LONGWORD_OUT:
                         set_wrapped_lw(cntx,get_item_bufadr(cntx, itmlst_item[j]),temp2[2*j]);
                         break;
                     case dk_vajhoej_vms_call_VMS_ITEM_QUADWORD_IN:
                         /* nothing */
                         break;
                     case dk_vajhoej_vms_call_VMS_ITEM_QUADWORD_OUT:
                         memcpy(&temp8,&temp2[2*j],sizeof(long long));
                         set_wrapped_qw(cntx,get_item_bufadr(cntx, itmlst_item[j]),temp8);
                         break;
                     case dk_vajhoej_vms_call_VMS_ITEM_CHARACTERSTRING_IN:
                         set_ba(cntx,itmlst_item_ba[j],itmlst_item_ba_dptr[j],0);
                         break;
                     case dk_vajhoej_vms_call_VMS_ITEM_CHARACTERSTRING_OUT:
                         set_ba(cntx,itmlst_item_ba[j],itmlst_item_ba_dptr[j],1);
                         break;
                 }
             }
#ifdef DEBUG
             for(j = 0; j <= nitem; j++)
             {
                 fprintf(stderr, "  %04X%04X\n", itmlst_arrstruct[j].code, itmlst_arrstruct[j].buflen);
                 if(itmlst_arrstruct[j].bufadr > 0x100)
                 {
                     p = (int *)itmlst_arrstruct[j].bufadr;
                     fprintf(stderr, "  %08X -> %08X%08X\n", itmlst_arrstruct[j].bufadr,  p[0], p[1]);
                 }
                 else
                 {
                     fprintf(stderr, "  %08X\n", itmlst_arrstruct[j].bufadr);
                 }
                 fprintf(stderr, "  %08X\n", itmlst_arrstruct[j].retlenadr);
             }
#endif
             break;
         case dk_vajhoej_vms_call_VMS_ROUTINE_0:
#ifdef DEBUG
            fprintf(stderr, "    r rbuf=%s\n", rbuf_ba_dptr->p32);
#endif
             set_ba(cntx,rbuf_ba,rbuf_ba_dptr,1);
             break;
      }
   }
   /* cleanup */
   (*cntx)->ReleaseIntArrayElements(cntx, argtyp, argtyp_int, 0);
   (*cntx)->ReleaseIntArrayElements(cntx, argacc, argacc_int, 0);
   free(arglist);
   free(argval_tw);
   free(argval_ba);
   free(argval_ba_dptr);
   free(argval_itmlst_arritm);
   free(itmlst_item);
   free(itmlst_typ);
   free(temp);
   free(itmlst_item_ba);
   free(itmlst_item_ba_dptr);
   free(itmlst_arrstruct);
   free(temp2);
   return result;
}

